#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <stdbool.h>


#define F_CPU 20000000UL
#include <util/delay.h>

#define xHallInPIN  PIND
#define xHallInPORT PORTD
#define xHallInDDR  DDRD
#define xHallInBit  2



#define xMagnetOutPORT PORTD
#define xMagnetOutDDR  DDRD
#define xMagnetOutBit  3


void xIOinit(){
  xMagnetOutDDR |= _BV(xMagnetOutBit) ;
  xMagnetOutPORT &= ~ _BV(xMagnetOutBit) ;
 
  xHallInDDR &= ~ _BV(xHallInBit) ;
  xHallInPORT |= _BV(xHallInBit) ;
  }

//---------------------------------------------------------------- 

void uartPutc( char c){
  while (!(UCSR0A & (1<<UDRE0))); 
  UDR0 = c;            
  }

/*
void uartPutc(char c){
   while (!(UCSR0A & (1<<UDRE0))); 
   UDR0 = c;                      
   }
*/
void uartPutsPgm(const char *progmem_s){
  register char c;
  while ( (c = pgm_read_byte(progmem_s++)) ) { 
    uartPutc(c); 
	}
  }

void exit(int k){
  uartPutsPgm(PSTR("exit")) ;
  for(;;){}
  }

void uartBlank(){
  uartPutc(' ') ; 
  }

void uartCrlf(){
  uartPutc(13) ; 
  uartPutc(10) ;
  }


void uartPutsPgmCR(const char *progmem_s){
  uartPutsPgm(progmem_s) ;
  uartCrlf() ;
  }



/*
void uartLf(){
  uartPutc(13) ; 
  }
*/

void uartHex(uint8_t v){
  v &=0b00001111 ;
  if (v<=9) { uartPutc('0'+v) ; } 
       else { uartPutc(v+'A'-10) ; } ;
  }

void uartByte(uint8_t v){
  uartHex(v>>4) ; 
  uartHex(v) ; 
  }

void uartWord(uint16_t v){
  uartByte(v>>8) ; 
  uartByte(v) ; 
  }

void uartDword(uint32_t v){
  uartWord(v>>16) ; 
  uartWord(v) ; 
  }


/*
void uartUint32(uint32_t v){
  uartWord(v>>16) ; 
  uartWord(v) ; 
  }
*/

void uartDecOut(int k){
  if (k<0) { k=-k ; uartPutc('-') ; } else { uartPutc(' ') ; }
  uartPutc( ((k/1000) % 10)+48 ) ;
  uartPutc( ((k/100) % 10)+48 ) ;
  uartPutc( ((k/10) % 10)+48 ) ;
  uartPutc( (k % 10)+48 ) ;
//   uartPutc(32) ;
  }

void uartDecOut32(int32_t k){
  if (k<0) { k=-k ; uartPutc('-') ; } else { uartPutc(' ') ; }
  uartPutc( ((k/10000000L) % 10)+48 ) ;
  uartPutc( ((k/1000000L) % 10)+48 ) ;   
  uartPutc( ((k/100000L) % 10)+48 ) ;  
  uartPutc( ((k/10000L) % 10)+48 ) ;
  uartPutc( ((k/1000L) % 10)+48 ) ;
  uartPutc( ((k/100L) % 10)+48 ) ;
  uartPutc( ((k/10L) % 10)+48 ) ;
  uartPutc( (k % 10L)+48 ) ;
//   uartPutc(32) ;
  }

// RS232 RX routines


uint8_t uartGetc(void){
 // while (!(UCSRA & (1<<RXC)));  
 // return UDR ;                   
  
  while (!(UCSR0A & (1<<RXC0)));  
  return UDR0 ;                   
  }

uint8_t uartGetcEcho(void){
  char c ;
  c=uartGetc() ;
  uartPutc(c) ;
  return(c) ;
  }

uint8_t uartChkc(void) {
  //return (UCSRA & (1<<RXC)) ;  //  Zeichen verfuegbar ?
   return (UCSR0A & (1<<RXC0)) ;  //  Zeichen verfuegbar ?
  }



void DelayMs(uint16_t x){
  uint16_t k ;
  for (k=0 ; k<x ; k++ ) { 
     _delay_ms(1) ;
    }
  }

//----------------------------------------------------------------


void autoRun(){
  while(1){
    xMagnetOutPORT |= _BV(xMagnetOutBit) ; 
	DelayMs(10) ;
    xMagnetOutPORT &= ~ _BV(xMagnetOutBit) ;
	DelayMs(40) ;
    }
  }


void timer1Init(){

  TCCR1A=0b00000010 ; // fast pwm no output
  TCCR1B=0b00011001	; // ck=XTAL
  // Interupt enable for Timer1-Overflow
  TIMSK1=0b00000001 ;
  OCR1A=1000 ;
  // ICR1=2000-1 ; // sample period predivider 
  ICR1=1600-1 ; // sample period predivider 
  }

volatile uint16_t timeCount1 ;

int Dt,T ;

ISR(TIMER1_OVF_vect) {
//   IO4port |=  _BV(IO4pin) ;
  timeCount1++ ;
  if(timeCount1==Dt){xMagnetOutPORT |= _BV(xMagnetOutBit) ;  }
  if(timeCount1==Dt+T){ xMagnetOutPORT &= ~ _BV(xMagnetOutBit) ;}
  if(timeCount1>2000){  timeCount1=0 ; }
//  IO4port &= ~ _BV(IO4pin) ;
  }

//----------------------------------------------------------------------------------------

uint8_t getHall(){
  return  (xHallInPIN & _BV(xHallInBit)) ==0 ;
  }

void measure1(){
  while(1){
    while( getHall() ){} ;
	while( !getHall()) {} ;
    cli() ;
    timeCount1=0 ;
    sei() ;
    while( getHall() ) {} ;
	while( !getHall() ){} ;
	cli() ;
    uint16_t tMeasure=timeCount1 ;
    sei() ;
    uartDecOut(tMeasure) ;
	uartCrlf() ;
	DelayMs(100) ;
	}
  }

int hallLast ;


void measureAndControl1(){
    cli() ;
	uint16_t tMeasure=timeCount1 ;
    timeCount1=0 ;
    sei() ;
   
//  20MHz
/*	int T1=400 ;
	if(tMeasure<500){ T1=320 ; } 
	if(tMeasure<400){ T1=220 ; } 
	if(tMeasure<300){ T1=180 ; } 
	if(tMeasure<250){ T1=140 ; } 
	if(tMeasure<200){ T1=100 ; } 
	if(tMeasure<180){ T1=80 ; } 
	if(tMeasure<160){ T1=60 ; } 
	if(tMeasure<140){ T1=40 ; } 
*/

	int T1=400 ;
	if(tMeasure<500){ T1=300 ; } 
	if(tMeasure<400){ T1=200 ; } 
	if(tMeasure<300){ T1=160 ; } 
	if(tMeasure<250){ T1=120 ; } 
	if(tMeasure<180){ T1=100 ; } 
	if(tMeasure<160){ T1=80 ; } 
	if(tMeasure<140){ T1=70 ; } 
	if(tMeasure<120){ T1=60 ; } 
	if(tMeasure<110){ T1=60 ; } 
	if(tMeasure<100){ T1=40 ; } 
	if(tMeasure<250){ T1=tMeasure-50 ; } 

	T=T1 ;
	uartDecOut(tMeasure) ;
	uartBlank() ;
    // tMeasure is in ms
	uartDecOut(60000UL/tMeasure) ;

	uartCrlf() ;
	}

void RunWithMeasure1(){
  Dt=5 ; T=100 ;
  Dt=8 ;
  Dt=10 ;
  while(1){
    int hall=getHall() ;
	if((hall==0) & (hallLast==1) ){ measureAndControl1() ; }
	//if((hall==1) & (hallLast==0) ){ measureAndControl1() ; }
    hallLast=hall ;
    }
  }

void directCoupleHallToMagnet(){
  Dt=30000 ; T=10 ;
  while(1){
    int hall=getHall() ;
	if( !hall ){
	  xMagnetOutPORT |= _BV(xMagnetOutBit) ; 
	  } 
	 else{ 
	  xMagnetOutPORT &= ~ _BV(xMagnetOutBit) ;
	  }
	if((hall==0) & (hallLast==1) ){ 
	  cli() ;
	  uint16_t tMeasure=timeCount1 ;
      timeCount1=0 ;
      sei() ;
      uartDecOut(tMeasure) ;
	  uartCrlf() ; 
	  }
    hallLast=hall ;
    }
  }

//----------------------------------------------------------------
int main(){

  DDRD =0b11011110 ; // enable outputs , bit 0,1=input
  PORTD=0b11111111 ; // pullups on on input pins
  
  DDRB =0b11111111 ; // enable outputs 
  PORTB=0b11111111 ; // pullups on on input pins





// 16.00MHz/19200/16=52
// 20MHz/16/115200=11
  UCSR0B |= ( 1 << TXEN0 ) | ( 1 << RXEN0 ) ;					// UART TX einschalten
  // UCSR0C |= ( 1 << URSEL )|( 3<<UCSZ0 );	    // Asynchron 8N1
  UBRR0H  = 0;                               // Highbyte ist 0
  //UBRR0L  = 11-1 ;                             // 115200 Bit/sec at 20MHz
  UBRR0L  = 9-1 ;                             // 115200 Bit/sec at 16MHz
  

  xIOinit() ;
  timer1Init() ;

  uartCrlf() ; 
   uartPutsPgm(PSTR("Start")) ;
   sei() ;

   RunWithMeasure1() ;
//   measure1() ;
  directCoupleHallToMagnet() ;
  while(1){
   
    } ;
  return(0) ;
 }


